"""Miscellaneous functions (https://mariadb.com/kb/en/miscellaneous-functions/)."""
from sqlfactory import Statement
from sqlfactory.func.base import Function


# pylint: disable=too-few-public-methods
class GetLock(Function):
    """Tries to obtain a lock with a name."""
    def __init__(self, name: str | Statement, timeout: int | Statement):
        super().__init__('GET_LOCK', name, timeout)


# pylint: disable=too-few-public-methods
class Inet6Aton(Function):
    """Converts an IPv6 address from its string representation to a binary string."""
    def __init__(self, ip: str | Statement):
        super().__init__('INET6_ATON', ip)


# pylint: disable=too-few-public-methods
class Inet6Ntoa(Function):
    """Converts an IPv6 address from its binary string representation to a string."""
    def __init__(self, ip: str | Statement):
        super().__init__('INET6_NTOA', ip)


# pylint: disable=too-few-public-methods
class InetAton(Function):
    """Converts an IPv4 address from its string representation to a number."""
    def __init__(self, ip: str | Statement):
        super().__init__('INET_ATON', ip)


# pylint: disable=too-few-public-methods
class InetNtoa(Function):
    """Converts an IP number to a string representation."""
    def __init__(self, ip: int | Statement):
        super().__init__('INET_NTOA', ip)


# pylint: disable=too-few-public-methods
class IsFreeLock(Function):
    """Checks whether a named lock is free."""
    def __init__(self, name: str | Statement):
        super().__init__('IS_FREE_LOCK', name)


# pylint: disable=too-few-public-methods
class IsIpv4(Function):
    """Checks whether a string is an IPv4 address."""
    def __init__(self, ip: str | Statement):
        super().__init__('IS_IPV4', ip)


# pylint: disable=too-few-public-methods
class IsIpv4Compat(Function):
    """Checks whether IPv6 address is a valid IPv4-compatible address."""
    def __init__(self, ip: str | Statement):
        super().__init__('IS_IPV4_COMPAT', ip)


# pylint: disable=too-few-public-methods
class IsIpv4Mapped(Function):
    """Checks whether IPv6 is an IPv4-mapped address."""
    def __init__(self, ip: str):
        super().__init__('IS_IPV4_MAPPED', ip)


# pylint: disable=too-few-public-methods
class IsIpv6(Function):
    """Checks whether a string is an IPv6 address."""
    def __init__(self, ip: str | Statement):
        super().__init__('IS_IPV6', ip)


# pylint: disable=too-few-public-methods
class IsUsedLock(Function):
    """Checks whether a named lock is in use."""
    def __init__(self, name: str | Statement):
        super().__init__('IS_USED_LOCK', name)


# pylint: disable=too-few-public-methods
class MasterGtidWait(Function):
    """Waits until the slave reaches a specified GTID position."""
    def __init__(self, gtid_set: str | Statement, timeout: int | Statement = None):
        if timeout is not None:
            super().__init__('MASTER_GTID_WAIT', gtid_set, timeout)
        else:
            super().__init__('MASTER_GTID_WAIT', gtid_set)


# pylint: disable=too-few-public-methods
class MasterPosWait(Function):
    """Waits until the slave reaches a specified binary log position."""
    def __init__(self, log_file: str | Statement, log_pos: int | Statement, timeout: int | Statement = None):
        if timeout is not None:
            super().__init__('MASTER_POS_WAIT', log_file, log_pos, timeout)
        else:
            super().__init__('MASTER_POS_WAIT', log_file, log_pos)


# pylint: disable=too-few-public-methods
class ReleaseAllLocks(Function):
    """Releases all named locks."""
    def __init__(self):
        super().__init__('RELEASE_ALL_LOCKS')


# pylint: disable=too-few-public-methods
class ReleaseLock(Function):
    """Releases a named lock."""
    def __init__(self, name: str | Statement):
        super().__init__('RELEASE_LOCK', name)


# pylint: disable=too-few-public-methods
class Sleep(Function):
    """Sleeps for a specified number of seconds."""
    def __init__(self, seconds: int | Statement):
        super().__init__('SLEEP', seconds)


# pylint: disable=too-few-public-methods
class SysGuid(Function):
    """Returns a globally unique identifier."""
    def __init__(self):
        super().__init__('SYS_GUID')


# pylint: disable=too-few-public-methods
class Uuid(Function):
    """Returns a universally unique identifier."""
    def __init__(self):
        super().__init__('UUID')


# pylint: disable=too-few-public-methods
class UuidShort(Function):
    """Returns a short universally unique identifier."""
    def __init__(self):
        super().__init__('UUID_SHORT')



"""SELECT statement builder."""

from .select import Select, SELECT
from .join import Join, LeftJoin, RightJoin, LeftOuterJoin, RightOuterJoin, InnerJoin, CrossJoin
from .column_list import ColumnList
from .aliased import SelectColumn, Aliased



"""Base classes for SQL functions"""

from typing import Any

from ..statement import StatementWithArgs, Statement


class Function(StatementWithArgs):
    """Generic function with name and variable number of arguments."""
    def __init__(self, function: str, *args: Statement | Any):
        self.function = function
        self._args = args

    def _args_placeholders(self) -> list[str]:
        out: list[str] = []

        for arg in self._args:
            if isinstance(arg, Statement):
                out.append(str(arg))
            else:
                out.append("%s")

        return out

    def __str__(self):
        return f"{self.function}({', '.join(self._args_placeholders())})"

    @property
    def args(self) -> list[Any]:
        out = []
        for arg in self._args:
            if isinstance(arg, StatementWithArgs):
                out.extend(arg.args)
            elif not isinstance(arg, Statement):
                out.append(arg)

        return out



"""Simple binary comparison conditions."""

from typing import Any

from .base import Condition, StatementOrColumn
from ..statement import Statement, StatementWithArgs


# pylint: disable=too-few-public-methods  # As everything is handled in base classes.
class SimpleCondition(Condition):
    # pylint: disable=duplicate-code  # It does not make sense to generalize two-row statement used on two places.
    """
    Simple condition comparing one column with given value, using specified operator.
    """
    def __init__(self, column: StatementOrColumn, operator: str, value: Statement | Any):
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param operator: Operator to use for comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if not isinstance(column, Statement):
            # pylint: disable=import-outside-toplevel,cyclic-import
            from ..entities import Column
            column = Column(column)

        args = []

        if isinstance(column, StatementWithArgs):
            args.extend(column.args)

        if isinstance(value, StatementWithArgs):
            args.extend(value.args)

        elif not isinstance(value, Statement):
            args.append(value)

        if isinstance(value, Statement):
            super().__init__(
                f"{str(column)} {operator} {str(value)}",
                *args
            )
        else:
            super().__init__(
                f"{str(column)} {operator} %s",
                *args
            )


# pylint: disable=too-few-public-methods  # As everything is handled in base classes.
class Equals(SimpleCondition):
    """
    - `column` = <value>
    - `column` IS NULL
    - <statement> = <value>
    - <statement> IS NULL
    """
    def __init__(self, column: StatementOrColumn, value: Any | None | Statement):
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if value is None:
            super().__init__(column, "IS", value)
        else:
            super().__init__(column, "=", value)


# pylint: disable=too-few-public-methods  # As everything is handled in base classes.
class NotEquals(SimpleCondition):
    """
    - `column` != <value>
    - <statement> != <value>
    - `column` IS NOT NULL
    - <statement> IS NOT NULL
    """
    def __init__(self, column: StatementOrColumn, value: Any | None | Statement):
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if value is None:
            super().__init__(column, "IS NOT", value)
        else:
            super().__init__(column, "!=", value)


# pylint: disable=too-few-public-methods  # As everything is handled in base classes.
class GreaterThanOrEquals(SimpleCondition):
    """
    - `column` >= <value>
    - <statement> >= <value>
    """
    def __init__(self, column: StatementOrColumn, value: Any | Statement):
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, ">=", value)


# pylint: disable=too-few-public-methods  # As everything is handled in base classes.
class GreaterThan(SimpleCondition):
    """
    - `column` > <value>
    - <statement> > <value>
    """
    def __init__(self, column: StatementOrColumn, value: Any | Statement):
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, ">", value)


# pylint: disable=too-few-public-methods  # As everything is handled in base classes.
class LessThanOrEquals(SimpleCondition):
    """
    - `column` <= <value>
    - <statement> <= <value>
    """
    def __init__(self, column: StatementOrColumn, value: Any | Statement):
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, "<=", value)


# pylint: disable=too-few-public-methods  # As everything is handled in base classes.
class LessThan(SimpleCondition):
    """
    - `column` < <value>
    - <statement> < <value>
    """
    def __init__(self, column: StatementOrColumn, value: Any | Statement):
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, "<", value)


# Convenient aliases for shorter code.
Eq = Equals
Ge = GreaterThanOrEquals
Gt = GreaterThan
Le = LessThanOrEquals
Lt = LessThan
Ne = NotEquals



"""WHERE mixin for query generator."""

from typing import Generic, TypeVar

from ..condition.base import ConditionBase

T = TypeVar("T")


class WithWhere(Generic[T]):
    """Mixin to provide WHERE support for query generator."""
    def __init__(self, *args, where: ConditionBase = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._where = where

    def where(self, condition: ConditionBase) -> T:
        """Set WHERE condition for query."""
        if self._where is not None:
            raise AttributeError("Where has already been specified.")

        self._where = condition
        return self

    # pylint: disable=invalid-name
    def WHERE(self, condition: ConditionBase) -> T:
        """Alias for where() to be more SQL-like with all capitals."""
        return self.where(condition)



"""LIKE statement"""

from typing import Any

from .base import Condition, StatementOrColumn
from ..entities import Column
from ..statement import Statement, StatementWithArgs


class Like(Condition):
    """
    SQL LIKE statement

    `column` LIKE %s
    `column` NOT LIKE %s
    <statement> LIKE %s
    <statement> NOT LIKE %s
    """
    def __init__(self, column: StatementOrColumn, value: Any | Statement, negative: bool = False):
        args = []

        if not isinstance(column, Statement):
            column = Column(column)

        if isinstance(column, StatementWithArgs):
            args.extend(column.args)

        if isinstance(value, StatementWithArgs):
            args.extend(value.args)
        else:
            args.append(value)

        if isinstance(value, Statement):
            super().__init__(
                f"{str(column)}{' NOT' if negative else ''} LIKE {str(value)}",
                *args,
            )
        else:
            super().__init__(
                f"{str(column)}{' NOT' if negative else ''} LIKE %s",
                *args
            )

    @staticmethod
    def escape(s: str) -> str:
        """
        Escape string for use in LIKE statement
        """
        return s.replace("%", "%%").replace("_", "__")



"""Date/Time SQL functions (https://mariadb.com/kb/en/date-time-functions/)."""
from typing import Any, Literal

from .base import Function
from .. import StatementWithArgs, Statement, Raw


# pylint: disable=too-many-instance-attributes
class Interval(StatementWithArgs):
    """
    INTERVAL statement for DATE_ADD, DATE_SUB and similar functions.
    """
    _ATTRIBUTES = (
        "microsecond", "second", "minute", "hour", "day", "week", "month", "quarter", "year",
        "second_microsecond", "minute_microsecond", "minute_second", "hour_microsecond", "hour_second",
        "hour_minute", "day_microsecond", "day_second", "day_minute", "day_hour", "year_month"
    )

    # pylint: disable=too-many-arguments, too-many-locals
    def __init__(
            self,
            microsecond: Any = None,
            second: Any = None,
            minute: Any = None,
            hour: Any = None,
            day: Any = None,
            week: Any = None,
            month: Any = None,
            quarter: Any = None,
            year: Any = None,
            second_microsecond: Any = None,
            minute_microsecond: Any = None,
            minute_second: Any = None,
            hour_microsecond: Any = None,
            hour_second: Any = None,
            hour_minute: Any = None,
            day_microsecond: Any = None,
            day_second: Any = None,
            day_minute: Any = None,
            day_hour: Any = None,
            year_month: Any = None
    ):
        self.microsecond = microsecond
        self.second = second
        self.minute = minute
        self.hour = hour
        self.day = day
        self.week = week
        self.month = month
        self.quarter = quarter
        self.year = year
        self.second_microsecond = second_microsecond
        self.minute_microsecond = minute_microsecond
        self.minute_second = minute_second
        self.hour_microsecond = hour_microsecond
        self.hour_second = hour_second
        self.hour_minute = hour_minute
        self.day_microsecond = day_microsecond
        self.day_second = day_second
        self.day_minute = day_minute
        self.day_hour = day_hour
        self.year_month = year_month

        if all(getattr(self, attr) is None for attr in self._ATTRIBUTES):
            raise ValueError("At least one attribute must be set.")

    def __str__(self):
        return "INTERVAL " + " ".join(
            f"{'%s' if not isinstance(getattr(self, attr), Statement) else str(getattr(self, attr))} {attr.upper()}"
            for attr in reversed(self._ATTRIBUTES)
            if getattr(self, attr) is not None
        )

    @property
    def args(self) -> list[Any]:
        args = []

        for attr in reversed(self._ATTRIBUTES):
            value = getattr(self, attr)
            if value is None:
                continue

            if isinstance(value, StatementWithArgs):
                args.extend(value.args)

            elif not isinstance(value, Statement):
                args.append(value)

        return args


# pylint: disable=too-few-public-methods
class AddMonths(Function):
    """
    Adds months to date.
    """
    def __init__(self, date: Any, months: int):
        super().__init__("ADDMONTHS", date, months)


# pylint: disable=too-few-public-methods
class AddDate(Function):
    """
    Adds days to date.
    """
    def __init__(self, date: Any, days: int | Interval):
        super().__init__("ADDDATE", date, days)


# pylint: disable=too-few-public-methods
class AddTime(Function):
    """
    Adds time to date.
    """
    def __init__(self, date: Any, time: Any):
        super().__init__("ADDTIME", date, time)


# pylint: disable=too-few-public-methods
class ConvertTz(Function):
    """
    Converts date from one timezone to another.
    """
    def __init__(self, date: Any, from_tz: str, to_tz: str):
        super().__init__("CONVERT_TZ", date, from_tz, to_tz)


# pylint: disable=too-few-public-methods
class CurDate(Function):
    """
    Returns current date.
    """
    def __init__(self):
        super().__init__("CURDATE")


# pylint: disable=too-few-public-methods
class CurrentDate(Function):
    """
    Returns current date.
    """
    def __init__(self):
        super().__init__("CURRENT_DATE")


# pylint: disable=too-few-public-methods
class CurrentTime(Function):
    """
    Returns current time.
    """
    def __init__(self):
        super().__init__("CURRENT_TIME")


# pylint: disable=too-few-public-methods
class CurrentTimestamp(Function):
    """
    Returns current timestamp.
    """
    def __init__(self):
        super().__init__("CURRENT_TIMESTAMP")


# pylint: disable=too-few-public-methods
class CurTime(Function):
    """
    Returns current time.
    """
    def __init__(self):
        super().__init__("CURTIME")


# pylint: disable=too-few-public-methods
class Date(Function):
    """
    Extracts date from datetime.
    """
    def __init__(self, date: Any):
        super().__init__("DATE", date)


# pylint: disable=too-few-public-methods
class DateDiff(Function):
    """
    Returns difference between two dates.
    """
    def __init__(self, date1: Any, date2: Any):
        super().__init__("DATEDIFF", date1, date2)


# pylint: disable=too-few-public-methods
class DateAdd(Function):
    """
    Adds interval to date.
    """
    def __init__(self, date: Any, interval: Interval):
        super().__init__("DATE_ADD", date, interval)


# pylint: disable=too-few-public-methods
class DateFormat(Function):
    """
    Formats date.
    """
    def __init__(self, date: Any, date_format: str):
        super().__init__("DATE_FORMAT", date, date_format)


# pylint: disable=too-few-public-methods
class DateSub(Function):
    """
    Subtracts interval from date.
    """
    def __init__(self, date: Any, interval: Interval):
        super().__init__("DATE_SUB", date, interval)


# pylint: disable=too-few-public-methods
class Day(Function):
    """
    Extracts day from date.
    """
    def __init__(self, date: Any):
        super().__init__("DAY", date)


# pylint: disable=too-few-public-methods
class DayName(Function):
    """
    Returns name of the day.
    """
    def __init__(self, date: Any):
        super().__init__("DAYNAME", date)


# pylint: disable=too-few-public-methods
class DayOfMonth(Function):
    """
    Extracts day of month from date.
    """
    def __init__(self, date: Any):
        super().__init__("DAYOFMONTH", date)


# pylint: disable=too-few-public-methods
class DayOfWeek(Function):
    """
    Extracts day of week from date.
    """
    def __init__(self, date: Any):
        super().__init__("DAYOFWEEK", date)


# pylint: disable=too-few-public-methods
class DayOfYear(Function):
    """
    Extracts day of year from date.
    """
    def __init__(self, date: Any):
        super().__init__("DAYOFYEAR", date)


# pylint: disable=too-few-public-methods
class Extract(Function):
    """
    Extracts part of date.
    """
    def __init__(self, unit: str, date: Any):
        args = [date] if not isinstance(date, Statement) else date.args if isinstance(date, StatementWithArgs) else []
        super().__init__(
            "EXTRACT",
            Raw(
                f"{unit} FROM {str(date) if isinstance(date, Statement) else '%s'}",
                *args
            )
        )


# pylint: disable=too-few-public-methods
class FormatPicoTime(Function):
    """
    Formats pico time.
    """
    def __init__(self, time: Any):
        super().__init__("FORMAT_PICO_TIME", time)


# pylint: disable=too-few-public-methods
class FromDays(Function):
    """
    Converts days to date.
    """
    def __init__(self, days: Any):
        super().__init__("FROM_DAYS", days)


# pylint: disable=too-few-public-methods
class FromUnixTime(Function):
    """
    Converts UNIX timestamp to date.
    """
    def __init__(self, unix_timestamp: Any):
        super().__init__("FROM_UNIXTIME", unix_timestamp)


# pylint: disable=too-few-public-methods
class GetFormat(Function):
    """
    Returns date format.
    """
    def __init__(
            self,
            date_format: Literal["DATE", "DATETIME", "TIME"],
            locale: Literal["EUR", "USA", "JIS", "ISO", "INTERNAL"]
    ):
        super().__init__("GET_FORMAT", Raw(date_format), locale)


# pylint: disable=too-few-public-methods
class Hour(Function):
    """
    Extracts hour from datetime.
    """
    def __init__(self, time: Any):
        super().__init__("HOUR", time)


# pylint: disable=too-few-public-methods
class LastDay(Function):
    """
    Returns last day of month.
    """
    def __init__(self, date: Any):
        super().__init__("LAST_DAY", date)


# pylint: disable=too-few-public-methods
class LocalTime(Function):
    """
    Returns local time.
    """
    def __init__(self):
        super().__init__("LOCALTIME")


# pylint: disable=too-few-public-methods
class LocalTimestamp(Function):
    """
    Returns local timestamp.
    """
    def __init__(self):
        super().__init__("LOCALTIMESTAMP")


# pylint: disable=too-few-public-methods
class MakeDate(Function):
    """
    Creates date from year and day of year.
    """
    def __init__(self, year: Any, day_of_year: Any):
        super().__init__("MAKEDATE", year, day_of_year)


# pylint: disable=too-few-public-methods
class MakeTime(Function):
    """
    Creates time from hours, minutes, seconds.
    """
    def __init__(self, hours: Any, minutes: Any, seconds: Any):
        super().__init__("MAKETIME", hours, minutes, seconds)


# pylint: disable=too-few-public-methods
class Microsecond(Function):
    """
    Extracts microseconds from datetime.
    """
    def __init__(self, time: Any):
        super().__init__("MICROSECOND", time)


# pylint: disable=too-few-public-methods
class Minute(Function):
    """
    Extracts minutes from datetime.
    """
    def __init__(self, time: Any):
        super().__init__("MINUTE", time)


# pylint: disable=too-few-public-methods
class MonthName(Function):
    """
    Returns name of the month.
    """
    def __init__(self, date: Any):
        super().__init__("MONTHNAME", date)


# pylint: disable=too-few-public-methods
class Now(Function):
    """
    Returns current date and time.
    """
    def __init__(self):
        super().__init__("NOW")


# pylint: disable=too-few-public-methods
class PeriodAdd(Function):
    """
    Add months to a period.
    """
    def __init__(self, date: Any, period: Any):
        super().__init__("PERIOD_ADD", date, period)


# pylint: disable=too-few-public-methods
class PeriodDiff(Function):
    """
    Returns number of months between two periods.
    """
    def __init__(self, period1: Any, period2: Any):
        super().__init__("PERIOD_DIFF", period1, period2)


# pylint: disable=too-few-public-methods
class Quarter(Function):
    """
    Extracts quarter from date.
    """
    def __init__(self, date: Any):
        super().__init__("QUARTER", date)


# pylint: disable=too-few-public-methods
class Second(Function):
    """
    Extracts seconds from datetime.
    """
    def __init__(self, time: Any):
        super().__init__("SECOND", time)


# pylint: disable=too-few-public-methods
class SecToTime(Function):
    """
    Converts seconds to time.
    """
    def __init__(self, seconds: Any):
        super().__init__("SEC_TO_TIME", seconds)


# pylint: disable=too-few-public-methods
class StrToDate(Function):
    """
    Converts string to date.
    """
    def __init__(self, date: Any, date_format: str):
        super().__init__("STR_TO_DATE", date, date_format)


# pylint: disable=too-few-public-methods
class SubDate(Function):
    """
    Subtracts days from date.
    """
    def __init__(self, date: Any, days: int | Interval):
        super().__init__("SUBDATE", date, days)


# pylint: disable=too-few-public-methods
class SubTime(Function):
    """
    Subtracts time from date.
    """
    def __init__(self, date: Any, time: Any):
        super().__init__("SUBTIME", date, time)


# pylint: disable=too-few-public-methods
class SysDate(Function):
    """
    Returns current date.
    """
    def __init__(self):
        super().__init__("SYSDATE")


# pylint: disable=too-few-public-methods
class Time(Function):
    """
    Extracts time from datetime.
    """
    def __init__(self, time: Any):
        super().__init__("TIME", time)



class TimeDiff(Function):
    """
    Returns difference between two times.
    """
    def __init__(self, time1: Any, time2: Any):
        super().__init__("TIMEDIFF", time1, time2)


# pylint: disable=too-few-public-methods
class Timestamp(Function):
    """
    Converts date to timestamp.
    """
    def __init__(self, date: Any):
        super().__init__("TIMESTAMP", date)


# pylint: disable=too-few-public-methods
class TimeFormat(Function):
    """
    Formats time.
    """
    def __init__(self, time: Any, date_format: str):
        super().__init__("TIME_FORMAT", time, date_format)


# pylint: disable=too-few-public-methods
class TimeToSec(Function):
    """
    Converts time to seconds.
    """
    def __init__(self, time: Any):
        super().__init__("TIME_TO_SEC", time)


# pylint: disable=too-few-public-methods
class ToDays(Function):
    """
    Converts date to days.
    """
    def __init__(self, date: Any):
        super().__init__("TO_DAYS", date)


# pylint: disable=too-few-public-methods
class ToSeconds(Function):
    """
    Converts date to seconds.
    """
    def __init__(self, date: Any):
        super().__init__("TO_SECONDS", date)


# pylint: disable=too-few-public-methods
class UnixTimestamp(Function):
    """
    Returns UNIX timestamp.
    """
    def __init__(self):
        super().__init__("UNIX_TIMESTAMP")


# pylint: disable=too-few-public-methods
class UtcDate(Function):
    """
    Returns current date in UTC.
    """
    def __init__(self):
        super().__init__("UTC_DATE")


# pylint: disable=too-few-public-methods
class UtcTime(Function):
    """
    Returns current time in UTC.
    """
    def __init__(self):
        super().__init__("UTC_TIME")


# pylint: disable=too-few-public-methods
class UtcTimestamp(Function):
    """
    Returns current timestamp in UTC.
    """
    def __init__(self):
        super().__init__("UTC_TIMESTAMP")


# pylint: disable=too-few-public-methods
class Week(Function):
    """
    Returns week number.
    """
    def __init__(self, date: Any, mode: int | Statement = None):
        if mode is not None:
            super().__init__("WEEK", date, mode)
        else:
            super().__init__("WEEK", date)


# pylint: disable=too-few-public-methods
class WeekDay(Function):
    """
    Returns day of week.
    """
    def __init__(self, date: Any):
        super().__init__("WEEKDAY", date)


# pylint: disable=too-few-public-methods
class WeekOfYear(Function):
    """
    Returns week of year.
    """
    def __init__(self, date: Any):
        super().__init__("WEEKOFYEAR", date)


# pylint: disable=too-few-public-methods
class Year(Function):
    """
    Extracts year from date.
    """
    def __init__(self, date: Any):
        super().__init__("YEAR", date)


# pylint: disable=too-few-public-methods
class YearWeek(Function):
    """
    Returns year and week number.
    """
    def __init__(self, date: Any, mode: int | Statement = None):
        if mode is None:
            super().__init__("YEARWEEK", date)
        else:
            super().__init__("YEARWEEK", date, mode)



"""Encryption functions (https://mariadb.com/kb/en/encryption-hashing-and-compression-functions/)"""

from typing import Any, Literal

from .base import Function
from .. import Statement


# pylint: disable=too-few-public-methods
class AesDecrypt(Function):
    """
    Decrypts an encrypted string using AES.
    """
    def __init__(self, value: Any, key: Any):
        super().__init__('AES_DECRYPT', value, key)


# pylint: disable=too-few-public-methods
class AesEncrypt(Function):
    """
    Encrypts a string using AES.
    """
    def __init__(self, value: Any, key: Any):
        super().__init__('AES_ENCRYPT', value, key)


# pylint: disable=too-few-public-methods
class Compress(Function):
    """
    Compress a string.
    """
    def __init__(self, value: Any):
        super().__init__('COMPRESS', value)


# pylint: disable=too-few-public-methods
class DesDecrypt(Function):
    """
    Decrypts an encrypted string using DES.
    """
    def __init__(self, value: Any, key: Any):
        super().__init__('DES_DECRYPT', value, key)


# pylint: disable=too-few-public-methods
class DesEncrypt(Function):
    """
    Encrypts a string using DES.
    """
    def __init__(self, value: Any, key: Any):
        super().__init__('DES_ENCRYPT', value, key)


# pylint: disable=too-few-public-methods
class Encode(Function):
    """
    Encode a string.
    """
    def __init__(self, value: Any, encoding: str):
        super().__init__('ENCODE', value, encoding)


# pylint: disable=too-few-public-methods
class Decode(Function):
    """
    Decode a string.
    """
    def __init__(self, value: Any, encoding: str):
        super().__init__('DECODE', value, encoding)


# pylint: disable=too-few-public-methods
class Encrypt(Function):
    """
    Encrypt a string using Unix crypt()
    """
    def __init__(self, value: Any, salt: Any = None):
        if salt is None:
            super().__init__('ENCRYPT', value)
        else:
            super().__init__('ENCRYPT', value, salt)


# pylint: disable=too-few-public-methods
class Kdf(Function):
    """
    Key derivation function.
    """
    # pylint: disable=too-many-arguments
    def __init__(
            self, key: Any, salt: Any, info_or_iterations: Any = None,
            kdf_name: Literal['hkdf', 'pbkdf2_hmac'] | Statement = None, width: Any = None
    ):
        args = [key, salt]

        if info_or_iterations is not None:
            args.append(info_or_iterations)

            if kdf_name is not None:
                args.append(kdf_name)

                if width is not None:
                    args.append(width)

        super().__init__('KDF', *args)


# pylint: disable=too-few-public-methods
class OldPassword(Function):
    """
    Hash a string using the old MySQL password hashing algorithm.
    """
    def __init__(self, value: Any):
        super().__init__('OLD_PASSWORD', value)


# pylint: disable=too-few-public-methods
class Password(Function):
    """
    Hash a string using the MySQL password hashing algorithm.
    """
    def __init__(self, value: Any):
        super().__init__('PASSWORD', value)


# pylint: disable=too-few-public-methods
class MD5(Function):
    """
    Calculate an MD5 128-bit checksum.
    """
    def __init__(self, value: Any):
        super().__init__('MD5', value)


# pylint: disable=too-few-public-methods
class RandomBytes(Function):
    """
    Generate a random byte string.
    """
    def __init__(self, length: Any):
        super().__init__('RANDOM_BYTES', length)


# pylint: disable=too-few-public-methods
class Sha1(Function):
    """
    Calculate an SHA-1 160-bit checksum.
    """
    def __init__(self, value: Any):
        super().__init__('SHA1', value)


# pylint: disable=too-few-public-methods
class Sha2(Function):
    """
    Calculate an SHA-2 checksum.
    """
    def __init__(self, value: Any, length: Any):
        super().__init__('SHA2', value, length)


# pylint: disable=too-few-public-methods
class Uncompress(Function):
    """
    Uncompress a compressed string.
    """
    def __init__(self, value: Any):
        super().__init__('UNCOMPRESS', value)


# pylint: disable=too-few-public-methods
class UncompressLength(Function):
    """
    Return the length of an uncompressed string.
    """
    def __init__(self, value: Any):
        super().__init__('UNCOMPRESS_LENGTH', value)



"""BETWEEN condition generator"""

from typing import Any

from .base import Condition, StatementOrColumn
from ..entities import Column
from ..statement import Statement, StatementWithArgs


# pylint: disable=too-few-public-methods  # As everything is handled by base classes.
class Between(Condition):
    # pylint: disable=duplicate-code  # It does not make sense to generalize two-row statement used on two places.
    """
    Provides generation for following syntax:

    - `column` BETWEEN <lower_bound> AND <upper_bound>
    - `column` NOT BETWEEN <lower_bound> AND <upper_bound>
    - <statement> BETWEEN <lower_bound> AND <upper_bound>
    - <statement> NOT BETWEEN <lower_bound> AND <upper_bound>

    """
    def __init__(
            self, column: StatementOrColumn, lower_bound: Any | Statement, upper_bound: Any | Statement, *,
            negative: bool = False
    ):
        lower_bound_s = "%s"
        upper_bound_s = "%s"

        if not isinstance(column, Statement):
            column = Column(column)

        args = []

        if isinstance(column, StatementWithArgs):
            args.extend(column.args)

        if isinstance(lower_bound, Statement):
            lower_bound_s = str(lower_bound)
            if isinstance(lower_bound, StatementWithArgs):
                args.extend(lower_bound.args)
        else:
            args.append(lower_bound)

        if isinstance(upper_bound, Statement):
            upper_bound_s = str(upper_bound)
            if isinstance(upper_bound, StatementWithArgs):
                args.extend(upper_bound.args)
        else:
            args.append(upper_bound)

        super().__init__(
            f"{str(column)} {'NOT ' if negative else ''}BETWEEN {lower_bound_s} AND {upper_bound_s}",
            *args
        )



"""Numeric SQL functions (https://mariadb.com/kb/en/numeric-functions/)"""
from typing import Any, overload

from sqlfactory.func.base import Function


# pylint: disable=too-few-public-methods
class Div(Function):
    """
    Integer division.
    """
    def __init__(self, dividend: Any, divisor: Any):
        super().__init__("DIV", dividend, divisor)


# pylint: disable=too-few-public-methods
class Abs(Function):
    """
    Absolute value.
    """
    def __init__(self, value: Any):
        super().__init__("ABS", value)


# pylint: disable=too-few-public-methods
class ACos(Function):
    """
    Arc cosine.
    """
    def __init__(self, value: Any):
        super().__init__("ACOS", value)


# pylint: disable=too-few-public-methods
class ASin(Function):
    """
    Arc sine.
    """
    def __init__(self, value: Any):
        super().__init__("ASIN", value)


# pylint: disable=too-few-public-methods
class ATan(Function):
    """
    Arc tangent.
    """
    def __init__(self, value: Any):
        super().__init__("ATAN", value)


# pylint: disable=too-few-public-methods
class ATan2(Function):
    """
    Arc tangent of two variables.
    """
    def __init__(self, y: Any, x: Any):
        super().__init__("ATAN2", y, x)


# pylint: disable=too-few-public-methods
class Ceil(Function):
    """
    Ceiling value.
    """
    def __init__(self, value: Any):
        super().__init__("CEIL", value)


# pylint: disable=too-few-public-methods
class Ceiling(Function):
    """
    Ceiling value.
    """
    def __init__(self, value: Any):
        super().__init__("CEILING", value)


# pylint: disable=too-few-public-methods
class Conv(Function):
    """
    Convert number from one base to another.
    """
    def __init__(self, number: Any, from_base: Any, to_base: Any):
        super().__init__("CONV", number, from_base, to_base)


# pylint: disable=too-few-public-methods
class Cos(Function):
    """
    Cosine.
    """
    def __init__(self, value: Any):
        super().__init__("COS", value)


# pylint: disable=too-few-public-methods
class Cot(Function):
    """
    Cotangent.
    """
    def __init__(self, value: Any):
        super().__init__("COT", value)


# pylint: disable=too-few-public-methods
class Crc32(Function):
    """
    Computes a cyclic redundancy check (CRC) value.
    """
    def __init__(self, value: Any):
        super().__init__("CRC32", value)


# pylint: disable=too-few-public-methods
class Crc32C(Function):
    """
    Computes a cyclic redundancy check (CRC) value using the Castagnoli polynomial
    """
    def __init__(self, value: Any):
        super().__init__("CRC32C", value)


# pylint: disable=too-few-public-methods
class Degrees(Function):
    """
    Converts radians to degrees.
    """
    def __init__(self, value: Any):
        super().__init__("DEGREES", value)


# pylint: disable=too-few-public-methods
class Exp(Function):
    """
    e raised to power of the argument
    """
    def __init__(self, value: Any):
        super().__init__("EXP", value)


# pylint: disable=too-few-public-methods
class Floor(Function):
    """
    Floor value.
    """
    def __init__(self, value: Any):
        super().__init__("FLOOR", value)


# pylint: disable=too-few-public-methods
class Greatest(Function):
    """
    Return the largest argument
    """
    def __init__(self, *args: Any):
        super().__init__("GREATEST", *args)


# pylint: disable=too-few-public-methods
class Least(Function):
    """
    Return the smallest argument
    """
    def __init__(self, *args: Any):
        super().__init__("LEAST", *args)


# pylint: disable=too-few-public-methods
class Ln(Function):
    """
    Natural logarithm
    """
    def __init__(self, value: Any):
        super().__init__("LN", value)


# pylint: disable=too-few-public-methods
class Log(Function):
    """
    Logarithm
    """
    @overload
    def __init__(self, value: Any): ...


    @overload
    def __init__(self, base: Any, value: Any): ...

    def __init__(self, arg1: Any, arg2: Any = None):
        if arg2 is not None:
            super().__init__("LOG", arg1, arg2)

        else:
            super().__init__("LOG", arg1)


# pylint: disable=too-few-public-methods
class Log10(Function):
    """
    Base-10 logarithm
    """
    def __init__(self, value: Any):
        super().__init__("LOG10", value)


# pylint: disable=too-few-public-methods
class Log2(Function):
    """
    Base-2 logarithm
    """
    def __init__(self, value: Any):
        super().__init__("LOG2", value)


# pylint: disable=too-few-public-methods
class Mod(Function):
    """
    Modulo operation
    """
    def __init__(self, dividend: Any, divisor: Any):
        super().__init__("MOD", dividend, divisor)


# pylint: disable=too-few-public-methods
class Oct(Function):
    """
    Convert number to octal
    """
    def __init__(self, value: Any):
        super().__init__("OCT", value)


# pylint: disable=too-few-public-methods
class Pi(Function):
    """
    Return the value of pi
    """
    def __init__(self):
        super().__init__("PI")


# pylint: disable=too-few-public-methods
class Pow(Function):
    """
    Return X raised to power of Y.
    """
    def __init__(self, x: Any, y: Any):
        super().__init__("POW", x, y)


# pylint: disable=too-few-public-methods
class Power(Function):
    """
    Return X raised to power of Y.
    """
    def __init__(self, x: Any, y: Any):
        super().__init__("POWER", x, y)


# pylint: disable=too-few-public-methods
class Radians(Function):
    """
    Converts degrees to radians.
    """
    def __init__(self, value: Any):
        super().__init__("RADIANS", value)


# pylint: disable=too-few-public-methods
class Rand(Function):
    """
    Return a random floating-point value.
    """
    def __init__(self):
        super().__init__("RAND")


# pylint: disable=too-few-public-methods
class Round(Function):
    """
    Round the argument to the nearest integer.
    """
    def __init__(self, value: Any):
        super().__init__("ROUND", value)


# pylint: disable=too-few-public-methods
class Sign(Function):
    """
    Sign of the argument
    """
    def __init__(self, value: Any):
        super().__init__("SIGN", value)


# pylint: disable=too-few-public-methods
class Sin(Function):
    """
    Sine
    """
    def __init__(self, value: Any):
        super().__init__("SIN", value)


# pylint: disable=too-few-public-methods
class Sqrt(Function):
    """
    Square root
    """
    def __init__(self, value: Any):
        super().__init__("SQRT", value)


# pylint: disable=too-few-public-methods
class Tan(Function):
    """
    Tangent
    """
    def __init__(self, value: Any):
        super().__init__("TAN", value)


# pylint: disable=too-few-public-methods
class Truncate(Function):
    """
    Truncate the argument to D decimal places.
    """
    def __init__(self, value: Any, d: Any):
        super().__init__("TRUNCATE", value, d)


class BitCount(Function):
    """
    Return the number of bits that are set in the argument.
    """
    def __init__(self, value: Any):
        super().__init__("BIT_COUNT", value)



"""ORDER BY mixin for query generator"""

from __future__ import annotations
from enum import Enum
from typing import Any, Generic, TypeVar, Collection, Literal

from ..entities import Column, ColumnArg
from ..statement import Statement, StatementWithArgs


class Direction(str, Enum):
    """Ordering direction as enum"""
    ASC = "ASC"
    DESC = "DESC"


OrderColumn = ColumnArg | Statement


class Order(list[tuple[OrderColumn, Direction | Literal['ASC', 'DESC']]], StatementWithArgs):
    """ORDER BY statement as list of columns to use for ordering"""
    def __str__(self):
        if not self:
            return ""

        out = []

        for column, direction in self:
            if isinstance(column, str):
                column = Column(column)

            out.append(f"{str(column)} {direction.value if isinstance(direction, Direction) else direction}")

        return f"ORDER BY {', '.join(out)}"

    @property
    def args(self) -> list[Any]:
        out = []
        for column, _ in self:
            if isinstance(column, StatementWithArgs):
                out.extend(column.args)

        return out


T = TypeVar('T')


class WithOrder(Generic[T]):
    """Mixin to provide ORDER BY support for query generator."""
    def __init__(self, *args, order: OrderArg = None, **kwargs):
        """
        :param order: Ordering specification - either instance of Order, or collection of columns and directions.
        """
        super().__init__(*args, **kwargs)
        if order:
            self._order = order if isinstance(order, Order) else Order(order)
        else:
            self._order = None

    def order_by(self, column: OrderColumn, direction: Direction) -> T:
        """
        Add column to be used for ordering. Can be called multiple times, columns will be ordered by order of calls.
        :param column: Column to use for ordering
        :param direction: Ordering direction
        :return:
        """
        if self._order is None:
            self._order = Order()

        self._order.append((column, direction))
        return self

    # pylint: disable=invalid-name
    def ORDER_BY(self, column: OrderColumn, direction: Direction) -> T:
        """Alias for order_by() to be more SQL-like with all capitals."""
        return self.order_by(column, direction)


# Ordering argument for class init (specify directly instance of Order, or collection of columns and directions).
OrderArg = Order | Collection[tuple[OrderColumn, Direction]]



"""DELETE statement builder"""

from typing import Any

from ..entities import Table
from ..condition.base import ConditionBase
from ..execute import ExecutableStatementWithArgs
from ..mixins.limit import WithLimit, Limit
from ..mixins.order import WithOrder, OrderArg
from ..mixins.where import WithWhere


# pylint: disable=too-many-ancestors  # This is intentional, as this class is a combination of multiple mixins.
class Delete(ExecutableStatementWithArgs, WithWhere['Delete'], WithOrder['Delete'], WithLimit['Delete']):
    """
    DELETE statement

    >>> Delete("table", where=In("id", [1, 2, 3]))
    >>> "DELETE FROM `table` WHERE `id` IN (1,2,3)"
    """
    def __init__(
            self,
            table: Table | str,
            where: ConditionBase = None,
            order: OrderArg = None,
            limit: Limit = None
    ):
        """
        :param table: Table to delete from
        :param where: WHERE condition
        :param order: Ordering of matched rows, usefull when limiting number of deleted rows.
        :param limit: Limit number of deleted rows.
        """
        super().__init__(where=where, order=order, limit=limit)
        self.table = table if isinstance(table, Table) else Table(table)

    def __str__(self):
        """Construct the DELETE statement."""
        q = [f"DELETE FROM {str(self.table)}"]

        if self._where:
            q.append("WHERE")
            q.append(str(self._where))

        if self._order:
            q.append(str(self._order))

        if self._limit:
            q.append(str(self._limit))

        return " ".join(q)

    @property
    def args(self) -> list[Any]:
        """DELETE statement arguments."""
        return (
            (self._where.args if self._where else []) +
            (self._order.args if self._order else []) +
            (self._limit.args if self._limit else [])
        )


# Alias to provide better SQL compatibility
DELETE = Delete



"""Information functions (https://mariadb.com/kb/en/information-functions/)."""
from typing import Any

from .base import Function
from .. import Statement, StatementWithArgs, Raw, Column


# pylint: disable=too-few-public-methods
class Benchmark(Function):
    """Executes an expression repeatedly."""
    def __init__(self, count: int, expression: Statement):
        super().__init__('BENCHMARK', count, expression)


# pylint: disable=too-few-public-methods
class BinlogGtidPos(Function):
    """Returns a string representation of the corresponding GTID position."""
    def __init__(self):
        super().__init__('BINLOG_GTID_POS')


# pylint: disable=too-few-public-methods
class Charset(Function):
    """Returns the character set."""
    def __init__(self):
        super().__init__('CHARSET')


# pylint: disable=too-few-public-methods
class Coercibility(Function):
    """Returns the collation coercibility value of the string expression."""
    def __init__(self, expression: str):
        super().__init__('COERCIBILITY', expression)


# pylint: disable=too-few-public-methods
class Collation(Function):
    """Collation of the string argument"""
    def __init__(self, expression: str):
        super().__init__('COLLATION', expression)


# pylint: disable=too-few-public-methods
class Collate(Raw):
    """String with collation"""
    def __init__(self, expression: str | Statement, collation: str):
        super().__init__(
            f"{str(expression) if isinstance(expression, Statement) else '%s'} COLLATE {collation}",
            *(
                expression.args if isinstance(expression, StatementWithArgs)
                else [expression] if not isinstance(expression, Statement)
                else []
            )
        )


# pylint: disable=too-few-public-methods
class ConnectionId(Function):
    """Connection ID"""
    def __init__(self):
        super().__init__('CONNECTION_ID')


# pylint: disable=too-few-public-methods
class CurrentRole(Function):
    """Current role name"""
    def __init__(self):
        super().__init__('CURRENT_ROLE')


# pylint: disable=too-few-public-methods
class CurrentUser(Function):
    """Username/host that authenticated the current client"""
    def __init__(self):
        super().__init__('CURRENT_USER')


# pylint: disable=too-few-public-methods
class Database(Function):
    """Current default database"""
    def __init__(self):
        super().__init__('DATABASE')


# pylint: disable=too-few-public-methods
class DecodeHistogram(Function):
    """Returns comma separated numerics corresponding to a probability distribution"""
    def __init__(self, hist_type: Any, histogram: Any):
        super().__init__('DECODE_HISTOGRAM', hist_type, histogram)


# pylint: disable=too-few-public-methods
class Default(Function):
    """Returns the default value for a table column"""
    def __init__(self, column: Column):
        super().__init__('DEFAULT', column)


# pylint: disable=too-few-public-methods
class FoundRows(Function):
    """Returns the number of (potentially) returned rows if there was no LIMIT involved."""
    def __init__(self):
        super().__init__('FOUND_ROWS')


# pylint: disable=too-few-public-methods
class LastInsertId(Function):
    """Returns the value generated for an AUTO_INCREMENT column by the previous INSERT statement."""
    def __init__(self):
        super().__init__('LAST_INSERT_ID')


# pylint: disable=too-few-public-methods
class LastValue(Function):
    """Evaluates expression and returns the last."""
    def __init__(self, expr: Statement, *exprs: Statement):
        super().__init__('LAST_VALUE', expr, *exprs)


# pylint: disable=too-few-public-methods
class RowNumber(Function):
    """Returns the number of accepted rows so far."""
    def __init__(self):
        super().__init__('ROW_NUMBER')


# pylint: disable=too-few-public-methods
class Schema(Function):
    """Current default schema"""
    def __init__(self):
        super().__init__('SCHEMA')


# pylint: disable=too-few-public-methods
class SessionUser(Function):
    """Username/host that authenticated the current client"""
    def __init__(self):
        super().__init__('SESSION_USER')


# pylint: disable=too-few-public-methods
class SystemUser(Function):
    """Username/host that authenticated the current client"""
    def __init__(self):
        super().__init__('SYSTEM_USER')


# pylint: disable=too-few-public-methods
class User(Function):
    """Username/host that authenticated the current client"""
    def __init__(self):
        super().__init__('USER')


# pylint: disable=too-few-public-methods
class Version(Function):
    """Database version"""
    def __init__(self):
        super().__init__('VERSION')



"""VALUES() function for usage in ON DUPLICATE KEY UPDATE statements. MySQL / MariaDB specific."""

from ..entities import ColumnArg, Column
from ..func.base import Function


# pylint: disable=too-few-public-methods  # Too few public methods, as everything is handled by super class.
class Values(Function):
    """
    VALUES(<column>) for usage in INSERT INTO ... ON DUPLICATE KEY UPDATE column = VALUES(column) statements.
    """
    def __init__(self, column: ColumnArg):
        if not isinstance(column, Column):
            column = Column(column)

        super().__init__("VALUES", column)



"""Base classes to implement SQL conditions."""

from __future__ import annotations

from abc import ABC
from collections.abc import Iterable
from typing import Any

from ..statement import StatementWithArgs, ConditionalStatement, Statement
StatementOrColumn = str | Statement


class ConditionBase(StatementWithArgs, ConditionalStatement, ABC):
    """
    Generic condition interface, that can be chained with other conditions using & or | operators. All condition
    classes should inherit from this one, as there are checks through the library for instances of this class.
    """
    def __and__(self, other: ConditionBase) -> And:
        if isinstance(self, And):
            out = self
        else:
            out = And(self)

        if isinstance(other, And):
            out.extend(other.sub_conditions)
        else:
            out.append(other)

        return out

    def __or__(self, other: ConditionBase) -> Or:
        if isinstance(self, Or):
            out = self
        else:
            out = Or(self)

        if isinstance(other, Or):
            out.extend(other.sub_conditions)
        else:
            out.append(other)

        return out


class Condition(ConditionBase):
    """
    Generic RAW condition with optional arguments.
    """

    def __init__(self, condition: str, *args: Any):
        """
        :param condition: Condition (such as "`column` = %s")
        :param args: Optional arguments used in condition.
        """
        self.condition = condition
        self._args = list(args)

    def __str__(self):
        return self.condition

    @property
    def args(self) -> list[Any]:
        return self._args

    def __bool__(self) -> bool:
        return bool(self.condition)


class CompoundCondition(ConditionBase):
    """
    Base class for joining multiple conditions together using specified operator. As there are only two operators
    (AND and OR), this class is not meant to be used directly, but rather through And and Or classes.
    """
    def __init__(self, operator: str, *conditions: ConditionBase | str):
        """
        :param operator: Which operator to use for joining specific conditions.
        :param conditions: Conditions to join using given operator.
        """
        self.operator = operator
        self._sub_conditions: list[ConditionBase] = [
            Condition(condition) if isinstance(condition, str) else condition
            for condition in conditions
        ]

    def append(self, condition: ConditionBase | str, *args: Any) -> CompoundCondition:
        """
        Append another condition to be joined.
        :param condition: Condition to append
        :param args: Optional arguments for given condition, if condition is passed as string.
        """
        if isinstance(condition, str):
            condition = Condition(condition, *args)
        elif args:
            raise AttributeError("*args can be used only for str conditions.")

        self._sub_conditions.append(condition)
        return self

    def extend(self, conditions: Iterable[ConditionBase]):
        """
        Extend condition with list of conditions.
        :param conditions: Conditions to extend this condition with
        """
        self._sub_conditions.extend(conditions)
        return self

    @property
    def sub_conditions(self) -> list[ConditionBase]:
        """
        Return filtered list of conditions, which are valid.
        """
        return list(filter(bool, self._sub_conditions))

    def __str__(self):
        """
        Create SQL statement of joined conditions.
        """
        sub_conditions = list(map(str, self.sub_conditions))
        if sub_conditions:
            return "(" + f" {self.operator} ".join(sub_conditions) + ")" if len(sub_conditions) > 1 else sub_conditions[0]

        return "TRUE"

    @property
    def args(self) -> list[Any]:
        """
        Return arguments from all sub conditions in correct order.
        """
        out = []

        for cond in self.sub_conditions:
            out.extend(cond.args)

        return out

    def __bool__(self):
        """
        Is this compound condition non-empty?
        """
        return bool(self.sub_conditions)


class And(CompoundCondition):
    """
    Compound condition joined by AND.

    Usage:
        >>> And(Equals("id", 1), Equals("name", "hello"))
        >>> "(`id` = 1 AND `name` = 'hello')"

    Of course, condition can be another compound condition:
        >>> And(Equals("id", 1), Or(Equals("name", "hello"), Equals("name", "world")))
        >>> "(`id` = 1 AND (`name` = 'hello' OR `name` = 'world'))"
    """
    def __init__(self, *conditions: ConditionBase | str):
        super().__init__("AND", *conditions)


class Or(CompoundCondition):
    """
    Compound condition joined by OR.

    Usage:
        >>> Or(Equals("id", 1), Equals("name", "hello"))
        >>> "(`id` = 1 OR `name` = 'hello')"
    """
    def __init__(self, *conditions: ConditionBase | str):
        super().__init__("OR", *conditions)



"""LIMIT statement"""

from typing import TypeVar, Generic, overload

from sqlfactory.statement import StatementWithArgs, ConditionalStatement

T = TypeVar("T")


class Limit(ConditionalStatement, StatementWithArgs):
    """LIMIT statement"""

    @overload
    def __init__(self):
        """No LIMIT statement"""

    @overload
    def __init__(self, limit: int):
        """
        Just a LIMIT statement without offset
        :param limit: Number of returned rows
        """

    @overload
    def __init__(self, offset: int, limit: int):
        """
        LIMIT statement with both offset and limit
        :param offset: Pagination offset (how many rows to skip before returning result)
        :param limit: Number of returned rows
        """

    def __init__(self, offset_or_limit: int = None, limit: int = None):
        """
        LIMIT statement
        :param offset_or_limit: Pagination offset, or limit if second argument is None
        :param limit: Number of returned rows.
        """

        if limit is None:
            limit = offset_or_limit
            offset_or_limit = None

        self.offset = offset_or_limit
        self.limit = limit

    def __str__(self) -> str:
        if self.offset is not None:
            return "LIMIT %s, %s"

        if self.limit is not None:
            return "LIMIT %s"

        return ""

    def __bool__(self) -> bool:
        """Return True if statement should be included in query, False otherwise."""
        return self.offset is not None or self.limit is not None

    @property
    def args(self) -> list[int]:
        if self.offset is not None:
            return [self.offset, self.limit]

        if self.limit is not None:
            return [self.limit]

        return []


class WithLimit(Generic[T]):
    """Mixin to provide LIMIT support for query generator."""
    def __init__(self, *args, limit: Limit = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._limit = limit

    @overload
    def limit(self, limit: Limit) -> T:
        """
        Limit statement
        :param limit: Instance of Limit
        """

    @overload
    def limit(self, limit: int) -> T:
        """
        Limit statement
        :param limit: Number of returned rows
        """

    @overload
    def limit(self, offset: int, limit: int) -> T:
        """
        Limit statement
        :param offset: Pagination offset (how many rows to skip before returning result)
        :param limit: Number of returned rows
        """

    def limit(self, offset_or_limit: int | Limit | None, limit: int = None) -> T:
        """Limit statement"""
        if self._limit is not None:
            raise AttributeError("Limit has already been specified.")

        if isinstance(offset_or_limit, Limit):
            self._limit = offset_or_limit

            if limit is not None:
                raise AttributeError(
                    "When passing Limit instance as first argument, second argument should not be passed.")

        else:
            self._limit = Limit(offset_or_limit, limit)

        return self

    # pylint: disable=invalid-name
    @overload
    def LIMIT(self, limit: Limit) -> T:
        """Alias for limit() to be more SQL-like with all capitals."""

    # pylint: disable=invalid-name
    @overload
    def LIMIT(self, limit: int) -> T:
        """Alias for limit() to be more SQL-like with all capitals."""

    # pylint: disable=invalid-name
    @overload
    def LIMIT(self, offset: int, limit: int) -> T:
        """Alias for limit() to be more SQL-like with all capitals."""

    # pylint: disable=invalid-name
    def LIMIT(self, offset_or_limit: int, limit: int = None) -> T:
        """Alias for limit() to be more SQL-like with all capitals."""
        return self.limit(offset_or_limit, limit)



"""Conditions for WHERE, ON, HAVING clauses in SQL statements."""

from .base import ConditionBase, Condition, And, Or
from .between import Between
from .in_condition import In
from .like import Like
from .simple import Equals, NotEquals, GreaterThan, GreaterThanOrEquals, LessThan, LessThanOrEquals, Eq, Ne, Gt, Ge, Lt, Le



"""JOIN statements for SQL queries."""

from typing import Any

from sqlfactory.entities import Table
from sqlfactory.condition.base import ConditionBase
from sqlfactory.statement import StatementWithArgs


class Join(StatementWithArgs):
    """JOIN statement"""
    def __init__(self, table: str | Table, on: ConditionBase = None, alias: str = None):
        if isinstance(table, str):
            table = Table(table)

        self.table = table
        self.on = on
        self.alias = alias

    @property
    def join_spec(self) -> str:
        """
        Returns the JOIN type itself for generation of SQL query.
        """
        return "JOIN"

    def __str__(self):
        if self.alias:
            table = f"{str(self.table)} AS `{self.alias}`"
        else:
            table = str(self.table)

        if self.on:
            return f"{self.join_spec} {table} ON {str(self.on)}"

        return f"{self.join_spec} {table}"

    @property
    def args(self) -> list[Any]:
        return self.on.args if self.on else []


class LeftJoin(Join):
    """LEFT JOIN statement"""

    @property
    def join_spec(self) -> str:
        return "LEFT JOIN"


class LeftOuterJoin(Join):
    """LEFT OUTER JOIN statement"""

    @property
    def join_spec(self) -> str:
        return "LEFT OUTER JOIN"


class RightJoin(Join):
    """RIGHT JOIN statement"""
    @property
    def join_spec(self) -> str:
        return "RIGHT JOIN"


class RightOuterJoin(Join):
    """RIGHT OUTER JOIN statement"""
    @property
    def join_spec(self) -> str:
        return "RIGHT OUTER JOIN"


class InnerJoin(Join):
    """INNER JOIN statement"""
    @property
    def join_spec(self) -> str:
        return "INNER JOIN"


class CrossJoin(Join):
    """CROSS JOIN statement"""
    @property
    def join_spec(self) -> str:
        return "CROSS JOIN"



"""INSERT statement builder."""

from __future__ import annotations

from collections.abc import Collection
from typing import Any

from .values import Values
from ..entities import ColumnArg, Column, Table
from ..execute import ExecutableStatementWithArgs, ConditionalExecutableStatement
from ..statement import Statement, StatementWithArgs


class Insert(ConditionalExecutableStatement, ExecutableStatementWithArgs):
    """
    INSERT statement

    Statement is conditional, which means it won't be executed if no rows would be inserted (which throws SQL error).

    >>> Insert.into("table")("column1", "column2", "column3").values((1, 2, 3), (4, 5, 6))
    >>> "INSERT INTO `table` (`column1`, `column2`, `column3`) VALUES (1, 2, 3), (4, 5, 6)"

    >>> Insert("table", ignore=True)("column1", "column2", "column3").values((1, 2, 3), (4, 5, 6))
    >>> "INSERT IGNORE INTO `table` (`column1`, `column2`, `column3`) VALUES (1, 2, 3), (4, 5, 6)"
    """

    def __init__(self, table: Table | str, ignore: bool = False, replace: bool = False):
        """
        :param table: Table to insert into
        :param ignore: use INSERT IGNORE?
        :param replace: use REPLACE?
        """

        super().__init__()

        if ignore and replace:
            raise AttributeError("Only one of ignore or replace can be specified.")

        self._ignore = ignore
        self._replace = replace

        self._table = table if isinstance(table, Table) else Table(table)
        self._columns: list[Column] = []
        self._values: list[Collection[Any]] = []
        self._on_duplicate_key_update_set: list[tuple[Column, str]] = []
        self._on_duplicate_key_update_args: list[Any] = []

    @classmethod
    def into(cls, table: Table | str, *, ignore: bool = False, replace: bool = False) -> Insert:
        """Specify table to insert into."""
        return cls(table, ignore=ignore, replace=replace)

    # pylint: disable=invalid-name
    @classmethod
    def INTO(cls, table: Table | str, *, ignore: bool = False, replace: bool = False) -> Insert:
        """Alias for into() to provide better SQL compatibility"""
        return cls.into(table, ignore=ignore, replace=replace)

    def __call__(self, *columns: ColumnArg):
        if not columns:
            raise AttributeError("At least one column must be specified.")

        if self._columns:
            raise AttributeError("Insert columns has already been specified.")

        for column in columns:
            if not isinstance(column, (Column, str)):
                raise AttributeError("Statements cannot be used as INSERT columns.")

        self._columns = [column if isinstance(column, Column) else Column(column) for column in columns]
        return self

    def values(self, *rows: Collection[Any]):
        """
        Specify values to insert. Each row should be one collection. The semantics is identical to the SQL syntax.

        >>> Insert.into("table")("a", "b").values(
        >>>    ("row 1, column a", "row 1, column b"),
        >>>    ("row 2, column a", "row 2, column b")
        >>> )
        """
        self._values.extend(rows)
        return self

    # pylint: disable=invalid-name
    def VALUES(self, *rows: Collection[Any]):
        """Alias for values() to provide better SQL compatibility."""
        return self.values(*rows)

    def on_duplicate_key_update(self, **kwargs: Values | Statement | Any):
        """
        MySQL / MariaDB specific. Specify columns to update if row already exists (duplicate key check is triggered).

        Specify individual columns to be updated as keyword arguments.

        You can use Values() function to access value from currently inserted row's values, e.g.:
        >>> Insert.into("table")("a", "b").values(
        >>>     (1, 2),
        >>>     (3, 4)
        >>> ).on_duplicate_key_update(
        >>>     a=Values("a"),          # Set column "a" to the value of column "a" from the row being inserted
        >>>     b=Column("b") + 1       # Increment value of column "b" by 1 each time the duplicate is detected.
        >>> )
        """
        for column, stmt in kwargs.items():
            if not isinstance(column, Column):
                column = Column(column)

            self._on_duplicate_key_update_set.append((column, str(stmt) if isinstance(stmt, Statement) else "%s"))
            if isinstance(stmt, StatementWithArgs):
                self._on_duplicate_key_update_args.extend(stmt.args)
            elif not isinstance(stmt, Statement):
                self._on_duplicate_key_update_args.append(stmt)

        return self

    # pylint: disable=invalid-name
    def ON_DUPLICATE_KEY_UPDATE(self, **kwargs: Values | Statement | Any):
        """Alias for on_duplicate_key_update() to provide better SQL compatibility."""
        return self.on_duplicate_key_update(**kwargs)

    def __bool__(self) -> bool:
        """Do not execute empty INSERT statement."""
        return bool(self._values)

    def __str__(self) -> str:
        """Constructs INSERT statement from provided data."""
        if not self._columns:
            raise AttributeError("At least one column must be specified.")

        if self._replace:
            q = [f"REPLACE INTO {str(self._table)}"]
        else:
            q = [f"INSERT{' IGNORE' if self._ignore else ''} INTO {str(self._table)}"]

        if self._columns:
            q.append(f"({', '.join(map(str, self._columns))})")

        q.append("VALUES")

        count_columns = len(self._columns)

        for idx, row in enumerate(self._values):
            if len(row) != count_columns:
                raise AttributeError(f"Row {idx} has different number of values than specified number of columns.")

            q.append(f"({', '.join(['%s'] * len(row))}){',' if idx < len(self._values) - 1 else ''}")

        if self._on_duplicate_key_update_set:
            q.append("ON DUPLICATE KEY UPDATE")
            q.append(", ".join(map(
                lambda update_set: f"{str(update_set[0])} = {update_set[1]}",
                self._on_duplicate_key_update_set
            )))

        return " ".join(q)

    @property
    def args(self) -> list[Any]:
        return [v for row in self._values for v in row] + self._on_duplicate_key_update_args


# Alias for Insert, for better SQL compatibility
INSERT = Insert



"""Column / Statement aliasing support"""

from typing import Any

from sqlfactory.entities import ColumnArg, Column
from sqlfactory.statement import StatementWithArgs, Statement


# pylint: disable=too-few-public-methods
class Aliased(StatementWithArgs):
    """Aliased generic statement. Only to be used in SELECT statement, where AS statement is only valid."""
    def __init__(self, statement: Statement | ColumnArg, alias: str = None):
        super().__init__()
        self._statement = statement if isinstance(statement, Statement) else Column(statement)
        self.alias = alias

    def __str__(self):
        if self.alias is None:
            return str(self._statement)

        return f"{str(self._statement)} AS `{self.alias}`"

    @property
    def args(self) -> list[Any]:
        return self._statement.args if isinstance(self._statement, StatementWithArgs) else []

    def __getattr__(self, name):
        """Proxy to access attributes of inner (non-aliased) statement."""
        return getattr(self._statement, name)


class SelectColumn(Aliased):
    """Aliased column"""
    def __init__(self, column: ColumnArg, alias: str = None):
        super().__init__(column, alias)



"""SELECT statement builder."""

from __future__ import annotations

from collections.abc import Collection
from typing import Any, overload

from .column_list import ColumnList
from .join import Join, LeftJoin
from ..entities import ColumnArg, Table
from ..condition.base import ConditionBase
from ..execute import ExecutableStatementWithArgs
from ..mixins.limit import WithLimit, Limit
from ..mixins.order import WithOrder, OrderArg
from ..mixins.where import WithWhere
from ..statement import StatementWithArgs, Statement


# pylint: disable=too-many-ancestors  # Intentional, as this class is a combination of multiple mixins.
class Select(ExecutableStatementWithArgs, WithWhere['Select'], WithOrder['Select'], WithLimit['Select']):
    # pylint: disable=too-many-arguments  # Yes, SELECT is complex.
    """
    SELECT statement

    >>> (Select("column1", "column2", "column3", table="table_name")
    >>>     .where(Eq("column1", 1))
    >>>     .order_by("column2")
    >>>     .limit(2, 10)
    >>>     .execute(cursor))

    Known limitations:
    - JOINs are not checked for uniqueness, so it is possible to add same JOIN multiple times. For now, it is up to
      user to ensure that JOINs are unique. This may be changed in the future, but it is not simple task, as joins
      needs to be in certain order.
    """
    def __init__(
            self,
            *columns: Statement | ColumnArg,
            select: ColumnList = None,
            table: Table | str | Statement = None,
            join: Collection[Join] = None,
            where: ConditionBase = None,
            group_by: ColumnList | Collection[Statement | ColumnArg] = None,
            having: ConditionBase = None,
            order: OrderArg = None,
            limit: Limit = None
    ):
        """
        :param columns: Columns to select.
        :param select: Columns to select as instance of ColumnList.
        :param table: Table to select from.
        :param join: Join statements
        :param where: Where condition
        :param group_by: Group by columns
        :param having: Having condition
        :param order: Order by columns
        :param limit: Limit results
        """
        super().__init__(where=where, order=order, limit=limit)

        if columns and select:
            raise AttributeError("Cannot specify individual columns when attribute select is present.")

        if select and not isinstance(select, ColumnList):
            raise TypeError("Select argument must be instance of ColumnList.")

        self.columns = select or ColumnList(columns)

        if not table:
            self.table = ""
            raise AttributeError("Missing required keyword argument table.")

        self.table = Table(table) if isinstance(table, str) else table
        self._join = list(join) if join is not None else None
        self._group_by = ColumnList(group_by) if group_by is not None and not isinstance(group_by, ColumnList) else group_by
        self._having = having

    def add(self, column: Statement | Any) -> Select:
        """Add new statement or column to the set of selected columns"""
        self.columns.add(column)
        return self

    def _append_join(self, join: Join) -> Select:
        """Append join to list of joins."""
        if not self._join:
            self._join = []

        if join not in self._join:
            self._join.append(join)

        return self

    @overload
    def join(self, join: Join) -> Select:
        """Append JOIN clause to the query (any Join instance)."""

    @overload
    def join(self, table: str | Table, on: ConditionBase = None, alias: str = None) -> Select:
        """Append JOIN clause to the query.
        JOIN `table` AS <alias> ON (<condition>) """

    def join(self, table: str | Table | Join, on: ConditionBase = None, alias: str = None) -> Select:
        """Append JOIN clause to the query.
        JOIN `table` AS <alias> ON (<condition>)
        """
        if isinstance(table, Join):
            if on is not None or alias is not None:
                raise AttributeError("When passing Join instance directly, on or alias attributes cannot be specified.")

            return self._append_join(table)

        return self._append_join(Join(table, on, alias))

    # pylint: disable=invalid-name
    def JOIN(self, table: str | Table | Join, on: ConditionBase = None, alias: str = None) -> Select:
        """Alias for join() to be more SQL-like with all capitals."""
        return self.join(table, on, alias)

    def left_join(self, table: str, on: ConditionBase = None, alias: str = None) -> Select:
        """Append LEFT JOIN clause to the query."""
        return self.join(LeftJoin(table, on, alias))

    # pylint: disable=invalid-name
    def LEFT_JOIN(self, table: str, on: ConditionBase = None, alias: str = None) -> Select:
        """Alias for left_join() to be more SQL-like with all capitals."""
        return self.left_join(table, on, alias)

    def group_by(self, column: Statement | ColumnArg, *columns: Statement | ColumnArg) -> Select:
        """
        GROUP BY clause.

        >>> Select().group_by("column1", "column2", "column3")
        """
        if self._group_by is not None:
            raise AttributeError("GROUP BY has already been specified.")

        self._group_by = ColumnList([column] + list(columns))
        return self

    # pylint: disable=invalid-name
    def GROUP_BY(self, column: Statement | ColumnArg, *columns: Statement | ColumnArg) -> Select:
        """Alias for group_by() to be more SQL-like with all capitals."""
        return self.group_by(column, *columns)

    def having(self, condition: ConditionBase) -> Select:
        """HAVING clause"""
        self._having = condition
        return self

    # pylint: disable=invalid-name
    def HAVING(self, condition: ConditionBase) -> Select:
        """Alias for having() to be more SQL-like with all capitals."""
        return self.having(condition)

    def __str__(self):
        out = ["SELECT", str(self.columns), f"FROM {str(self.table)}"]

        if self._join:
            out.extend(map(str, self._join))

        if self._where:
            out.append("WHERE")
            out.append(str(self._where))

        if self._group_by:
            out.append("GROUP BY")
            out.append(str(self._group_by))

        if self._having:
            out.append("HAVING")
            out.append(str(self._having))

        if self._order:
            out.append(str(self._order))

        if self._limit:
            out.append(str(self._limit))

        return " ".join(out)

    @property
    def args(self) -> list[Any]:
        out = self.columns.args

        if self._join:
            for join in self._join:
                out.extend(join.args)

        return (
            out +
            (self.table.args if isinstance(self.table, StatementWithArgs) else []) +
            (self._where.args if self._where else []) +
            (self._group_by.args if self._group_by else []) +
            (self._having.args if self._having else []) +
            (self._order.args if self._order else []) +
            (self._limit.args if self._limit else [])
        )


SELECT = Select



"""Control flow functions"""

from typing import Any

from ..statement import Statement
from .base import Function


# pylint: disable=too-few-public-methods
class IfNull(Function):
    """If expr1 is not NULL, IFNULL() returns expr1; otherwise it returns expr2."""
    def __init__(self, expr1: Statement | Any, expr2: Statement | Any):
        super().__init__("IFNULL", expr1, expr2)


# pylint: disable=too-few-public-methods
class NullIf(Function):
    """Returns NULL if expr1 = expr2 is true, otherwise returns expr1."""
    def __init__(self, expr1: Statement | Any, expr2: Statement | Any):
        super().__init__("NULLIF", expr1, expr2)


# pylint: disable=too-few-public-methods
class If(Function):
    """If expr1 is TRUE (expr1 <> 0 and expr1 <> NULL) then IF() returns expr2; otherwise it returns expr3."""
    def __init__(self, expr: Statement | Any, if_true: Statement | Any, if_false: Statement | Any):
        super().__init__("IF", expr, if_true, if_false)



"""Column list for usage in SELECT statement."""

from __future__ import annotations
from typing import Iterable, Any

from sqlfactory.entities import ColumnArg, Column
from sqlfactory.statement import StatementWithArgs, Statement


class ColumnList(StatementWithArgs, list[Statement]):
    """
    Unique(ish) set of columns to be used in SELECT statement.
    """
    def __init__(self, iterable: Iterable[Statement | ColumnArg] = None):
        if iterable:
            super().__init__(map(
                lambda i: Column(i) if not isinstance(i, Statement) else i,
                iterable
            ))
        else:
            super().__init__()

    def __contains__(self, other: Statement):
        """This needs custom implementation over default list.__contains__ because we need to compare Column objects,
        which would generate Eq() instances instead of doing comparison."""
        for item in self:
            if isinstance(item, Column) and isinstance(other, Column):
                if str(item) == str(other):
                    return True
            elif item == other:
                return True

        return False

    def add(self, element: Statement | str) -> ColumnList:
        """Add new columns to the set."""
        return self.append(element)

    def append(self, element: Statement | str) -> ColumnList:
        """Add new columns to the set."""
        if not isinstance(element, Statement):
            element = Column(element)

        if element not in self:
            super().append(element)

        return self

    def update(self, iterable: Iterable[Statement | str]) -> ColumnList:
        """Add multiple new columns to the set."""
        for item in iterable:
            self.add(item)

        return self

    def __str__(self):
        return ", ".join(map(str, self))

    def __repr__(self):
        return "[" + ", ".join(map(repr, self)) + "]"

    @property
    def args(self) -> list[Any]:
        out = []

        for item in self:
            if isinstance(item, StatementWithArgs):
                out.extend(item.args)

        return out



"""Mixins are not directly exported from this module, as it should not be needed to use them in user code.
Classes exported here, however, should be used as their instances are expected as mixin arguments."""

from .limit import Limit
from .order import Order, Direction



"""Base classes for building SQL statements."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any


class Statement(ABC):
    """
    Base class of serializable SQL statement. This class cannot hold any data, it is just an interface
    for other classes to implement.

    Everything that should be serialized to SQL should inherit from this class or it's subclasses, as many
    checks in code are based on instance of this class.

    To retrieve SQL statement, call str() on the instance.
    """

    @abstractmethod
    def __str__(self) -> str:
        """Return SQL statement representing the statement."""

    def __repr__(self):
        """Representation for debugging purposes, it is not used in SQL generation."""
        return self.__str__()

    def __hash__(self):
        """Return hash of this statement to be able to use it in unique collections."""
        return hash(str(self))

    def __eq__(self, other: 'Statement'):
        """Compares this statement to other."""
        if not isinstance(other, Statement):
            return False

        return str(self) == str(other)


class StatementWithArgs(Statement):
    """
    Base class of serializable SQL statement with arguments that should be escaped. This class cannot hold any data,
    it is just an interface for other classes to implement.

    Any class that needs SQL argument substitution should inherit from this class, as there are checks that tests
    whether passed object is instance of this class.

    To retrieve SQL statement, call `str()` on the instance. To retrieve arguments, access `args` property.
    """

    @property
    @abstractmethod
    def args(self) -> list[Any]:
        """Return arguments representing `%s` placeholders in statement returned by `__str__()`. Number of items
        in returned list must match number of `%s` placeholders in string returned by calling `__str__()`."""

    def __hash__(self):
        """Return hash of this statement to be able to use it in unique collections."""
        return super().__hash__() + sum(map(hash, self.args))

    def __eq__(self, other: 'StatementWithArgs'):
        """Compares this statement to other."""

        # If other is not instance of StatementWithArgs (it is basic statement probably) and we have no args,
        # we can compare it as normal statement.
        if not isinstance(other, StatementWithArgs):
            if not self.args:
                return super().__eq__(other)

            # If we have args and other does not, they are not equal.
            return False

        return self.args == other.args and super().__eq__(other)

    def __repr__(self):
        """Representation of statement including the arguments."""
        args = list(map(repr, self.args))
        if args:
            return f"{super().__repr__()} with arguments [{', '.join(args)}]"

        return super().__repr__()


# pylint: disable=too-few-public-methods  # As this is just an interface.
class ConditionalStatement(ABC):
    """
    Mixin that provides conditional execution of the statement (query will be executed only if statement is valid).

    This class is used for example for INSERT statements, to not execute empty INSERT. Or to not execute UPDATE
    if there are no columns to be updated.
    """
    @abstractmethod
    def __bool__(self) -> bool:
        """
        Return True if the statement should be executed.
        """


class Raw(StatementWithArgs):
    """
    RAW string statement (with optional args), that won't be processed in any way.
    """

    def __init__(self, sql: str, *args: Any):
        super().__init__()
        self._statement = sql
        self._args = args

    def __str__(self):
        return self._statement

    @property
    def args(self) -> list[Any]:
        return list(self._args)



"""Main SQLBuilder module. All public classes are exported from here."""

from .condition import *
from .delete import *
from .insert import *
from .mixins import *
from .select import *
from .update import *

from .entities import Table, Column
from .statement import Statement, StatementWithArgs, Raw



"""Classes representing database entities."""

from __future__ import annotations
from typing import Any

from .statement import Statement, Raw, StatementWithArgs
from .condition.simple import Gt, Ge, Eq, Lt, Le, Ne


class Column(Statement):
    """
    Column (optionally with table and database) as statement.

    Provides shorthand for creating conditional SQL statements and arithmetic SQL statements. You can use Column
    instance to directly create condition using simple operators (==, !=, <, <=, >, >=) or arithmetic operations
    (+, -, *, /, %).

    >>> Column("table.column") == 5
    >>> # Produces Eq(Column("table.column"), 5)

    >>> Column("table.column") + 5
    >>> # Produces Raw("`table`.`column` + %s", 5)
    """
    def __init__(self, column: str):
        super().__init__()

        self._column = column.split(".")
        if len(self._column) > 3:
            raise ValueError("Invalid column name (contains more than <database>.<table>.<column>).")

    def __str__(self) -> str:
        return ".".join(map(lambda x: f"`{x}`" if not x.startswith("`") else x, self._column))

    @property
    def column(self) -> str:
        """Returns column part of the column name."""
        return self._column[-1]

    @property
    def table(self) -> str | None:
        """Returns table part of the column name, if specified. If column specification does not contain table name,
        returns None."""
        try:
            return self._column[-2]
        except IndexError:
            return None

    @property
    def database(self) -> str | None:
        """Returns database part of the column name, if specified. If column specification does not contain database
        name, returns None."""
        try:
            return self._column[-3]
        except IndexError:
            return None

    def __hash__(self):
        return hash(str(self))

    def __gt__(self, other: Statement | Any) -> Gt:
        """Shorthand to produce conditional SQL statement <column> > <other>."""
        return Gt(self, other)

    def __ge__(self, other: Statement | Any) -> Ge:
        """Shorthand to produce conditional SQL statement <column> >= <other>."""
        return Ge(self, other)

    def __lt__(self, other: Statement | Any) -> Lt:
        """Shorthand to produce conditional SQL statement <column> < <other>."""
        return Lt(self, other)

    def __le__(self, other: Statement | Any) -> Le:
        """Shorthand to produce conditional SQL statement <column> <= <other>."""
        return Le(self, other)

    def __eq__(self, other: Statement | Any) -> Eq:
        """Shorthand to produce conditional SQL statement <column> = <other>."""
        return Eq(self, other)

    def __ne__(self, other: Statement | Any) -> Ne:
        """Shorthand to produce conditional SQL statement <column> != <other>."""
        return Ne(self, other)

    def __add__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} + {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __sub__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} - {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __mul__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} * {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __truediv__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} / {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __mod__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} % {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __and__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} & {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __or__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} | {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __xor__(self, other) -> StatementWithArgs:
        return Raw(
            f"{str(self)} ^ {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __lshift__(self, other):
        return Raw(
            f"{str(self)} << {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __rshift__(self, other):
        return Raw(
            f"{str(self)} >> {str(other) if isinstance(other, Statement) else '%s'}",
            *other.args if isinstance(other, StatementWithArgs)
            else [other] if not isinstance(other, Statement)
            else []
        )

    def __neg__(self):
        return Raw(f"~{str(self)}")


class Table(Statement):
    """
    Table (optionally with database) as statement

    By accessing Table's undefined attributes, instance of Column is returned. This allows to easily access columns of
    that table, and reference them in SQL statements.

    >>> Table("database.table").column_name == 5
    >>> # Produces Eq(Column("database.table.column_name"), 5)
    """

    def __init__(self, table: str):
        """
        :param table: Table name (optionally with database in form <database>.<table>). If database is not specified,
            it is assumed that table is in default database and SQL constructed will not contain any database name.
        """
        super().__init__()
        self._table = table.split(".")

        if len(self._table) > 2:
            raise ValueError("Invalid table name (contains more than <database>.<table>).")

    def __str__(self):
        return ".".join(map(lambda x: f"`{x}`" if not x.startswith("`") else x, self._table))

    @property
    def table(self) -> str | None:
        """Returns table part of the table name"""
        return self._table[-1]

    @property
    def database(self) -> str | None:
        """Returns database part of the table name, if specified. If table specification does not contain database,
        returns None."""
        try:
            return self._table[-2]
        except IndexError:
            return None

    def __getattr__(self, name: str) -> Column:
        """Returns column of that table."""
        return Column(f"{'.'.join(self._table)}.{name}")


# Alias for column that can be passed as instance of column or as string.
ColumnArg = Column | str



"""String functions (https://mariadb.com/kb/en/string-functions/)"""

from typing import Any

from ..statement import Statement
from .base import Function


# pylint: disable=too-few-public-methods
class Ascii(Function):
    """Numeric ASCII value of leftmost character."""
    def __init__(self, arg: Statement | Any):
        super().__init__("ASCII", arg)


# pylint: disable=too-few-public-methods
class Bin(Function):
    """Returns binary value"""
    def __init__(self, num: Statement | Any):
        super().__init__("BIN", num)


# pylint: disable=too-few-public-methods
class BitLength(Function):
    """Returns the length of a string in bits"""
    def __init__(self, s: Statement | Any):
        super().__init__("BIT_LENGTH", s)


# pylint: disable=too-few-public-methods
class Char(Function):
    """Returns string based on the integer values for the individual characters."""
    def __init__(self, *n: Statement | int):
        super().__init__("CHAR", *n)


# pylint: disable=too-few-public-methods
class CharLength(Function):
    """Length of the string in characters."""
    def __init__(self, s: Statement | Any):
        super().__init__("CHAR_LENGTH", s)


# pylint: disable=too-few-public-methods
class Chr(Function):
    """Returns string based on integer values of the individual characters."""
    def __init__(self, n: Statement | int):
        super().__init__("CHR", n)


# pylint: disable=too-few-public-methods
class Concat(Function):
    """Returns concatenated string"""
    def __init__(self, *s: Statement | Any):
        super().__init__("CONCAT", *s)


# pylint: disable=too-few-public-methods
class ConcatWs(Function):
    """Concatenate with separator"""
    def __init__(self, separator: Statement | Any, *s: Statement | Any):
        super().__init__("CONCAT_WS", separator, *s)


# pylint: disable=too-few-public-methods
class Hex(Function):
    """Returns a hexadecimal string representation of a decimal or string value."""
    def __init__(self, n: Statement | Any):
        super().__init__("HEX", n)


# pylint: disable=too-few-public-methods
class InStr(Function):
    """Returns the position of the first occurrence of substring in string"""
    def __init__(self, s: Statement | Any, substring: Statement | Any):
        super().__init__("INSTR", s, substring)


# pylint: disable=too-few-public-methods
class Left(Function):
    """Returns the leftmost number of characters"""
    def __init__(self, s: Statement | Any, n: Statement | int):
        super().__init__("LEFT", s, n)


# pylint: disable=too-few-public-methods
class Length(Function):
    """Returns the length of a string"""
    def __init__(self, s: Statement | Any):
        super().__init__("LENGTH", s)


# pylint: disable=too-few-public-methods
class Locate(Function):
    """Returns the position of the first occurrence of substring in string"""
    def __init__(self, substring: Statement | Any, s: Statement | Any):
        super().__init__("LOCATE", substring, s)


# pylint: disable=too-few-public-methods
class Lower(Function):
    """Converts a string to lower-case"""
    def __init__(self, s: Statement | Any):
        super().__init__("LOWER", s)


# pylint: disable=too-few-public-methods
class Lpad(Function):
    """Left-pad a string with another string"""
    def __init__(self, s: Statement | Any, n: Statement | int, pad: Statement | Any):
        super().__init__("LPAD", s, n, pad)


# pylint: disable=too-few-public-methods
class Ltrim(Function):
    """Removes leading spaces"""
    def __init__(self, s: Statement | Any):
        super().__init__("LTRIM", s)


# pylint: disable=too-few-public-methods
class Mid(Function):
    """Returns a substring"""
    def __init__(self, s: Statement | Any, start: Statement | int, length: Statement | int):
        super().__init__("MID", s, start, length)


# pylint: disable=too-few-public-methods
class OctetLength(Function):
    """Returns the length of a string in bytes"""
    def __init__(self, s: Statement | Any):
        super().__init__("OCTET_LENGTH", s)


# pylint: disable=too-few-public-methods
class Ord(Function):
    """Numeric value of leftmost character"""
    def __init__(self, s: Statement | Any):
        super().__init__("ORD", s)


# pylint: disable=too-few-public-methods
class Repeat(Function):
    """Repeat a string the specified number of times"""
    def __init__(self, s: Statement | Any, n: Statement | int):
        super().__init__("REPEAT", s, n)


# pylint: disable=too-few-public-methods
class Replace(Function):
    """Replace occurrences of a specified string"""
    def __init__(self, s: Statement | Any, from_: Statement | Any, to: Statement | Any):
        super().__init__("REPLACE", s, from_, to)


# pylint: disable=too-few-public-methods
class Reverse(Function):
    """Reverse a string"""
    def __init__(self, s: Statement | Any):
        super().__init__("REVERSE", s)


# pylint: disable=too-few-public-methods
class Right(Function):
    """Returns the rightmost number of characters"""
    def __init__(self, s: Statement | Any, n: Statement | int):
        super().__init__("RIGHT", s, n)


# pylint: disable=too-few-public-methods
class RPad(Function):
    """Right-pad a string with another string"""
    def __init__(self, s: Statement | Any, n: Statement | int, pad: Statement | Any):
        super().__init__("RPAD", s, n, pad)


# pylint: disable=too-few-public-methods
class RTrim(Function):
    """Removes trailing spaces"""
    def __init__(self, s: Statement | Any):
        super().__init__("RTRIM", s)


# pylint: disable=too-few-public-methods
class SFormat(Function):
    """Format a string"""
    def __init__(self, format_: Statement | Any, *args: Statement | Any):
        super().__init__("SFORMAT", format_, *args)


# pylint: disable=too-few-public-methods
class Space(Function):
    """Returns a string of spaces"""
    def __init__(self, n: Statement | int):
        super().__init__("SPACE", n)


# pylint: disable=too-few-public-methods
class Substr(Function):
    """Returns a substring"""
    def __init__(self, s: Statement | Any, start: Statement | int, length: Statement | int):
        super().__init__("SUBSTR", s, start, length)


# pylint: disable=too-few-public-methods
class Substring(Function):
    """Returns a substring"""
    def __init__(self, s: Statement | Any, start: Statement | int, length: Statement | int):
        super().__init__("SUBSTRING", s, start, length)


# pylint: disable=too-few-public-methods
class SubstringIndex(Function):
    """Returns a substring"""
    def __init__(self, s: Statement | Any, delimiter: Statement | Any, count: Statement | int):
        super().__init__("SUBSTRING_INDEX", s, delimiter, count)


# pylint: disable=too-few-public-methods
class ToBase64(Function):
    """Converts a string to base64"""
    def __init__(self, s: Statement | Any):
        super().__init__("TO_BASE64", s)


# pylint: disable=too-few-public-methods
class ToChar(Function):
    """Converts a date/time/timestamp type expression to a string"""
    def __init__(self, n: Statement | Any):
        super().__init__("TO_CHAR", n)


# pylint: disable=too-few-public-methods
class Trim(Function):
    """Removes leading and trailing spaces"""
    def __init__(self, s: Statement | Any):
        super().__init__("TRIM", s)


# pylint: disable=too-few-public-methods
class Unhex(Function):
    """Converts a hexadecimal pairs of digits to the character represented by the number."""
    def __init__(self, s: Statement | Any):
        super().__init__("UNHEX", s)


# pylint: disable=too-few-public-methods
class Upper(Function):
    """Converts a string to upper-case"""
    def __init__(self, s: Statement | Any):
        super().__init__("UPPER", s)



"""Aggregate functions."""

from typing import Literal

from .base import Function
from ..entities import Column, ColumnArg


# pylint: disable=too-few-public-methods
class AggregateFunction(Function):
    """Base class for aggregate functions"""
    def __init__(self, agg: str, column: ColumnArg):
        super().__init__(agg, Column(column) if isinstance(column, str) else column)


# pylint: disable=too-few-public-methods
class Avg(AggregateFunction):
    """AVG(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("AVG", column)


# pylint: disable=too-few-public-methods
class BitAnd(AggregateFunction):
    """BIT_AND(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("BIT_AND", column)


# pylint: disable=too-few-public-methods
class BitOr(AggregateFunction):
    """BIT_OR(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("BIT_OR", column)


# pylint: disable=too-few-public-methods
class BitXor(AggregateFunction):
    """BIT_XOR(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("BIT_XOR", column)


# pylint: disable=too-few-public-methods
class Count(Function):
    """COUNT(<column>)"""
    def __init__(self, column: ColumnArg | Literal['*']):
        super().__init__(
            "COUNT",
            column if isinstance(column, Column) or column == '*' else Column(column)
        )


# pylint: disable=too-few-public-methods
class Max(AggregateFunction):
    """MAX(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("MAX", column)


# pylint: disable=too-few-public-methods
class Min(AggregateFunction):
    """MIN(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("MIN", column)


# pylint: disable=too-few-public-methods
class Std(AggregateFunction):
    """STD(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("STD", column)


# pylint: disable=too-few-public-methods
class Sum(AggregateFunction):
    """SUM(<column>)"""
    def __init__(self, column: ColumnArg):
        super().__init__("SUM", column)



"""UPDATE statement builder."""

from __future__ import annotations
from typing import Any, Optional

from ..entities import ColumnArg, Column, Table
from ..execute import ExecutableStatementWithArgs, ConditionalExecutableStatement
from ..mixins.limit import WithLimit, Limit
from ..mixins.where import WithWhere
from ..statement import StatementWithArgs, Statement
from ..condition.base import Condition


class UpdateColumn(StatementWithArgs):
    """
    Represents one field that should be updated.
    """
    def __init__(self, column: ColumnArg, value: Statement | Any):
        self._column = column if isinstance(column, Column) else Column(column)
        self._value = value

    def __str__(self):
        return f"{str(self._column)} = {str(self._value) if isinstance(self._value, Statement) else '%s'}"

    def __hash__(self):
        return hash(self._column)

    def __eq__(self, other: UpdateColumn):
        if not isinstance(other, UpdateColumn):
            return False

        return str(self._column) == str(other._column)

    @property
    def args(self) -> list[Any]:
        """
        Return arguments for the update statement.
        """
        if isinstance(self._value, StatementWithArgs):
            return self._value.args

        if not isinstance(self._value, Statement):
            return [self._value]

        return []


# pylint: disable=too-many-ancestors  # This is intentional, as this class is a combination of multiple mixins.
class Update(ExecutableStatementWithArgs, ConditionalExecutableStatement, WithWhere['Update'], WithLimit['Update']):
    """
    Builds UPDATE statement SQL query.

    This is conditional SQL statement, so you can check whether it should be executed (would update any columns)
    by calling bool() on it. Also, if you are using execute() method of the statement, the execution won't be
    performed if bool() returns False.

    >>> Update("table").set("column1", 1).where(Eq("column2", 2))

    >>> t = Table("table")
    >>> Update(t).set(t.column1, 1).where(t.column2 == 2)
    """
    def __init__(
            self,
            table: Table | str,
            *fields: UpdateColumn,
            where: Optional[Condition] = None,
            limit: Optional[Limit] = None
    ):
        """
        :param table: Table to update.
        :param fields: List of UpdateColumn instances containing columns to be updated. This is not very pleasant way
            to create the statement, use set() method instead.
        :param where: WHERE condition
        :param limit: Limit number of updated rows
        """
        super().__init__(where=where, limit=limit)
        self.table = table if isinstance(table, Table) else Table(table)
        self.fields: list[UpdateColumn] = list(fields)

    def __str__(self) -> str:
        """
        Return the UPDATE statement with %s placeholders for arguments.
        """
        if not self.fields:
            raise AttributeError("At least one column must be updated.")

        query = [
            f"UPDATE {str(self.table)}",
            f"SET {', '.join(map(str, self.fields))}"
        ]

        if self._where:
            query.append("WHERE")
            query.append(str(self._where))

        if self._limit:
            query.append(str(self._limit))

        return " ".join(query)

    @property
    def args(self) -> list[Any]:
        """
        Return all arguments used in the UPDATE statement.
        """
        out = []

        for field in self.fields:
            out.extend(field.args)

        if self._where is not None:
            out.extend(self._where.args)

        if self._limit is not None:
            out.extend(self._limit.args)

        return out

    def __bool__(self):
        """
        Return True if there are any fields to be updated and the statement should be executed.
        """
        return bool(self.fields)

    def append(self, field: UpdateColumn):
        """
        Append new UpdateField to this UPDATE statement. Can be used when set() method is not sufficient.
        """
        if field in self.fields:
            raise AttributeError(f"Field '{field}' is already in the list of fields to be updated.")

        self.fields.append(field)
        return self

    def set(self, field: ColumnArg, value: Statement | Any):
        """
        Syntactical sugar for creating simple SET UpdateFields.
        :param field: Field name (without quotes).
        :param value: Value to set the field to (will be escaped).
        :return:
        """
        return self.append(UpdateColumn(field, value))

    # pylint: disable=invalid-name
    def SET(self, field: str, value: Any):
        """Alias for set() for better SQL compatibility"""
        return self.set(field, value)


# Alias for Update to provide better SQL compatibility
UPDATE = Update



"""IN condition, used for checking whether column value is in given list of values."""

from collections.abc import Collection
from typing import overload, Any

from ..entities import Column
from ..statement import StatementWithArgs, Statement
from .base import Condition, StatementOrColumn


# pylint: disable=too-few-public-methods   # Everything is handled by super classes.
class In(Condition):
    """
    IN condition:

    - `column` IN (%s, %s, %s)
    - <statement> IN (%s, %s, %s)

    OR

    - (`column1`, `column2`) IN ((%s, %s), (%s, %s), (%s, %s))
    - (<statement>, `column`) IN ((%s, %s), (%s, %s), (%s, %s))
    - (<statement>, <statement>) IN ((%s, %s), (%s, %s), (%s, %s))
    - (`column`, <statement>) IN ((%s, %s), (%s, %s), (%s, %s))

    Also supports comparing to None for single-column conditions (In("column", [1,2,3,None]) will work as expected).
    """

    @overload
    def __init__(
            self, columns: tuple[StatementOrColumn, ...], values: Collection[tuple[Any, ...]], *,
            negative: bool = False
    ):
        """Provides type definition for statement (`column1`, `column2`) IN ((%s, %s), (%s, %s), (%s, %s))"""

    @overload
    def __init__(self, column: StatementOrColumn, values: Collection[Any], *, negative: bool = False):
        """Provides type definition for statement `column` IN (%s, %s, %s)"""

    # pylint: disable=consider-using-f-string, too-many-branches  # Yes, IN statement is rather complex.
    def __init__(
            self, column: StatementOrColumn | tuple[StatementOrColumn, ...],
            values: Collection[Any | tuple[Any, ...]], *, negative: bool = False
    ):
        """
        :param column: Column to compare, or tuple of columns for multi-column comparison.
        :param values: Values to compare (list of values, or list of tuples of values).
        :param negative: Whether to perform negative comparison (NOT IN)
        """
        add_none = False

        if None in values:
            add_none = True
            values = list(filter(lambda v: v is not None, values))

        args = []

        if isinstance(column, tuple):
            column = tuple(Column(col) if not isinstance(col, Statement) else col for col in column)
        elif not isinstance(column, Statement):
            column = Column(column)

        if values:
            if isinstance(column, tuple):
                for stmt in column:
                    if isinstance(stmt, StatementWithArgs):
                        args.extend(stmt.args)

                for value_tuple in values:
                    for value in value_tuple:
                        if not isinstance(value, Statement):
                            args.append(value)
                        elif isinstance(value, StatementWithArgs):
                            args.extend(value.args)

                super().__init__(
                    "({}) {} ({})".format(
                        ", ".join(map(str, column)),
                        "IN" if not negative else "NOT IN ",
                        ", ".join(["(" + ", ".join([
                            "%s" if not isinstance(value, Statement) else str(value)
                            for value in value_tuple
                        ]) + ")" for value_tuple in values])
                    ),
                    *args,
                )
            else:
                in_stmt = "{} {} ({})".format(
                    str(column),
                    "IN" if not negative else "NOT IN",
                    ", ".join(["%s" if not isinstance(value, Statement) else str(value) for value in values])
                )

                if isinstance(column, StatementWithArgs):
                    args.extend(column.args)

                for value in values:
                    if isinstance(value, StatementWithArgs):
                        args.extend(value.args)
                    elif not isinstance(value, Statement):
                        args.append(value)

                if add_none:
                    if isinstance(column, StatementWithArgs):
                        args.extend(column.args)

                    super().__init__(
                        f"({in_stmt} {'OR' if not negative else 'AND'} {str(column)} IS {'NOT ' if negative else ''}NULL)",
                        *args
                    )
                else:
                    super().__init__(
                        in_stmt,
                        *args
                    )
        elif add_none:
            # This could happen only if there is just a one column, not multi-column statement.
            if isinstance(column, StatementWithArgs):
                args.extend(column.args)

            super().__init__(
                f"{str(column)} IS {'NOT ' if negative else ''}NULL",
                *args
            )
        else:
            super().__init__("FALSE" if not negative else "TRUE")



"""Module containing definition of executable SQL statements base classes."""

import inspect
from abc import ABC
from typing import Protocol, Any, overload

from .logger import logger
from .statement import Statement, StatementWithArgs, ConditionalStatement


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasQueryWithTupleArgs(Protocol):
    """Protocol defining DB driver with query method that takes arguments as tuple."""
    def query(self, query: str, args: tuple[Any]): ...


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasExecuteWithTupleArgs(Protocol):
    """Protocol defining DB driver with execute method that takes arguments as tuple."""
    def execute(self, query: str, args: tuple[Any]): ...


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasQueryWithArgs(Protocol):
    """Protocol defining DB driver with query method that takes arguments as multiple arguments."""
    def query(self, query: str, *args: Any): ...


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasExecuteWithArgs(Protocol):
    """Protocol defining DB driver with execute method that takes arguments as tuple."""
    def execute(self, query: str, args: tuple[Any]): ...


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasAsyncQueryWithTupleArgs(Protocol):
    """Protocol defining DB driver with async query method that takes arguments as tuple."""
    async def query(self, query: str, args: tuple[Any]): ...


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasAsyncExecuteWithTupleArgs(Protocol):
    """Protocol defining DB driver with async execute method that takes arguments as tuple."""
    async def execute(self, query: str, args: tuple[Any]): ...


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasAsyncQueryWithArgs(Protocol):
    """Protocol defining DB driver with async query method that takes arguments as multiple arguments."""
    async def query(self, query: str, *args: Any): ...


# pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
class HasAsyncExecuteWithArgs(Protocol):
    """Protocol defining DB driver with async execute method that takes arguments as tuple."""
    async def execute(self, query: str, args: tuple[Any]): ...


HasQueryOrExecute = HasQueryWithTupleArgs | HasExecuteWithTupleArgs | HasQueryWithArgs | HasExecuteWithArgs
HasAsyncQueryOrExecute = (
        HasAsyncQueryWithTupleArgs | HasAsyncExecuteWithTupleArgs | HasAsyncQueryWithArgs | HasAsyncExecuteWithArgs
)
MaybeAsyncHasQueryOrExecute = HasQueryOrExecute | HasAsyncQueryOrExecute


class ExecutableStatement(Statement, ABC):
    """
    This is the base class for an executable SQL statement that does not have any arguments.

    This class implements the execute() method. When given a database driver (or cursor) with query() or execute()
    methods, which take an SQL statement as the first argument and then a tuple or variadic arguments following SQL
    argument, it can be used to directly execute the SQL statement. This saves some typing by avoiding the manual
    passing of a string statement and arguments to the query.

    DB-API 2.0 drivers/cursors should all work with this implementation, as cursors should have execute() methods with
    the described semantics.

    Even async is supported. As this class does not process the result of the SQL statement in any way, the return
    value of the driver's execute()/query() method is directly returned. That returned value can be awaitable for async
    methods, so you can directly await it.
    """
    @overload
    def execute(self, trx: HasQueryOrExecute):
        """Execute statement on sync db driver."""

    @overload
    async def execute(self, trx: HasAsyncQueryOrExecute):
        """Execute statement on async db driver"""

    def execute(self, trx: MaybeAsyncHasQueryOrExecute, *args):
        """
        Execute statement on db driver (db-agnostic, just expects method `query` or `execute` on given driver).
        This is just a shortland for calling driver.execute(str(self), *args).
        :param trx: DB driver with query() or execute() method, which accepts either tuple as arguments,
         or multiple arguments following the query.
        :param args: Arguments to pass to the driver's query/execute method.
        :return: The same as db driver's execute/query method. If driver is async, returns awaitable response.
        """
        if hasattr(trx, "query"):
            call = trx.query
        elif hasattr(trx, "execute"):
            call = trx.execute
        else:
            raise AttributeError("trx must define query() or execute() method.")

        sig = inspect.signature(call)
        if any(map(lambda p: p.kind == inspect.Parameter.VAR_POSITIONAL, sig.parameters.values())):
            return call(str(self), *args)

        return call(str(self), tuple(args))


class ExecutableStatementWithArgs(StatementWithArgs, ExecutableStatement, ABC):
    """
    Base class of executable SQL statement with arguments that should be escaped.
    """
    def execute(self, trx: MaybeAsyncHasQueryOrExecute, *args):
        """
        Execute statement on db driver (db-agnostic, just expects method `query` or `execute` on given driver).
        Passes staement arguments to the driver's method. Returns response of the driver's method.
        This is just a shortland for calling driver.execute(str(self), *self.args).

        :param trx: DB driver with query() or execute() method, which accepts either tuple as arguments,
         or multiple arguments following the query.
        :param args: Additional arguments to append to the end of argument list (you probably don't need this).
        :return: The same as db driver's execute/query method. If driver is async, returns awaitable response.
        """
        return super().execute(trx, *self.args, *args)  # type: ignore


class ConditionalExecutableStatement(ExecutableStatement, ConditionalStatement, ABC):
    """
    Mixin that provides conditional execution of the statement (query will be executed only if statement is valid).

    This class is used for example for INSERT statements, to not execute empty INSERT. Or to not execute UPDATE
    if there are no columns to be updated.
    """
    def execute(self, trx: MaybeAsyncHasQueryOrExecute, *args):
        """
        Execute SQL statement using provided db-driver, but only if statement evaluates as True.
        :param trx: DB driver with query() or execute() method, which accepts either tuple as arguments,
        :param args: Arguments to pass to the db driver.
        :return:
        """
        if bool(self):
            return super().execute(trx, *args)  # type: ignore

        if inspect.iscoroutinefunction(getattr(trx, "query", getattr(trx, "execute", None))):
            logger.debug("Not executing statement, because it is false.")

            # pylint: disable=import-outside-toplevel
            import asyncio
            fut = asyncio.get_running_loop().create_future()
            fut.set_result(False)
            return fut

        logger.debug("Not executing statement, because it is false.")
        return False
